@@ -14,21 +14,22 @@ from typing import List
 from typing import Optional
 from typing import Union
 
-
 from ._compat import PY2
 from ._compat import PY38
+from ._compat import MutableMapping
 from ._compat import decode
 from ._compat import long
 from ._compat import unicode
 from ._utils import escape_string
 
+
 if PY2:
     from functools32 import lru_cache
 else:
     from functools import lru_cache
 
 
-def item(value, _parent=None):
+def item(value, _parent=None, _sort_keys=False):
     from .container import Container
 
     if isinstance(value, Item):
@@ -42,8 +43,11 @@ def item(value, _parent=None):
         return Float(value, Trivia(), str(value))
     elif isinstance(value, dict):
         val = Table(Container(), Trivia(), False)
-        for k, v in sorted(value.items(), key=lambda i: (isinstance(i[1], dict), i[0])):
-            val[k] = item(v, _parent=val)
+        for k, v in sorted(
+            value.items(),
+            key=lambda i: (isinstance(i[1], dict), i[0] if _sort_keys else 1),
+        ):
+            val[k] = item(v, _parent=val, _sort_keys=_sort_keys)
 
         return val
     elif isinstance(value, list):
@@ -57,13 +61,14 @@ def item(value, _parent=None):
                 table = Table(Container(), Trivia(), True)
 
                 for k, _v in sorted(
-                    v.items(), key=lambda i: (isinstance(i[1], dict), i[0])
+                    v.items(),
+                    key=lambda i: (isinstance(i[1], dict), i[0] if _sort_keys else 1),
                 ):
-                    i = item(_v)
+                    i = item(_v, _sort_keys=_sort_keys)
                     if isinstance(table, InlineTable):
                         i.trivia.trail = ""
 
-                    table[k] = item(i)
+                    table[k] = item(i, _sort_keys=_sort_keys)
 
                 v = table
 
@@ -201,7 +206,9 @@ class Key:
     A key value.
     """
 
-    def __init__(self, k, t=None, sep=None, dotted=False):  # type: (str) -> None
+    def __init__(
+        self, k, t=None, sep=None, dotted=False, original=None
+    ):  # type: (str, Optional[KeyType], Optional[str], bool, Optional[str]) -> None
         if t is None:
             if any(
                 [c not in string.ascii_letters + string.digits + "-" + "_" for c in k]
@@ -216,6 +223,11 @@ class Key:
 
         self.sep = sep
         self.key = k
+        if original is None:
+            original = k
+
+        self._original = original
+
         self._dotted = dotted
 
     @property
@@ -225,8 +237,11 @@ class Key:
     def is_dotted(self):  # type: () -> bool
         return self._dotted
 
+    def is_bare(self):  # type: () -> bool
+        return self.t == KeyType.Bare
+
     def as_string(self):  # type: () -> str
-        return "{}{}{}".format(self.delimiter, self.key, self.delimiter)
+        return "{}{}{}".format(self.delimiter, self._original, self.delimiter)
 
     def __hash__(self):  # type: () -> int
         return hash(self.key)
@@ -291,6 +306,9 @@ class Item(object):
     def is_inline_table(self):  # type: () -> bool
         return isinstance(self, InlineTable)
 
+    def is_aot(self):  # type: () -> bool
+        return isinstance(self, AoT)
+
     def _getstate(self, protocol=3):
         return (self._trivia,)
 
@@ -526,6 +544,12 @@ class Bool(Item):
 
         return other == self._value
 
+    def __hash__(self):
+        return hash(self._value)
+
+    def __repr__(self):
+        return repr(self._value)
+
 
 class DateTime(Item, datetime):
     """
@@ -545,7 +569,7 @@ class DateTime(Item, datetime):
         trivia,
         raw,
         **kwargs
-    ):  # type: (int, int, int, int, int, int, int, ..., Trivia, ...) -> datetime
+    ):  # type: (int, int, int, int, int, int, int, Optional[datetime.tzinfo], Trivia, str, Any) -> datetime
         return datetime.__new__(
             cls,
             year,
@@ -561,7 +585,7 @@ class DateTime(Item, datetime):
 
     def __init__(
         self, year, month, day, hour, minute, second, microsecond, tzinfo, trivia, raw
-    ):  # type: (int, int, int, int, int, int, int, ..., Trivia) -> None
+    ):  # type: (int, int, int, int, int, int, int, Optional[datetime.tzinfo], Trivia, str) -> None
         super(DateTime, self).__init__(trivia)
 
         self._raw = raw
@@ -650,7 +674,7 @@ class Date(Item, date):
     A date literal.
     """
 
-    def __new__(cls, year, month, day, *_):  # type: (int, int, int, ...) -> date
+    def __new__(cls, year, month, day, *_):  # type: (int, int, int, Any) -> date
         return date.__new__(cls, year, month, day)
 
     def __init__(
@@ -706,12 +730,12 @@ class Time(Item, time):
 
     def __new__(
         cls, hour, minute, second, microsecond, tzinfo, *_
-    ):  # type: (int, int, int, int, ...) -> time
+    ):  # type: (int, int, int, int, Optional[datetime.tzinfo], Any) -> time
         return time.__new__(cls, hour, minute, second, microsecond, tzinfo)
 
     def __init__(
         self, hour, minute, second, microsecond, tzinfo, trivia, raw
-    ):  # type: (int, int, int, int, Trivia, str) -> None
+    ):  # type: (int, int, int, int, Optional[datetime.tzinfo], Trivia, str) -> None
         super(Time, self).__init__(trivia)
 
         self._raw = raw
@@ -744,7 +768,9 @@ class Array(Item, list):
     An array literal
     """
 
-    def __init__(self, value, trivia, multiline=False):  # type: (list, Trivia) -> None
+    def __init__(
+        self, value, trivia, multiline=False
+    ):  # type: (list, Trivia, bool) -> None
         super(Array, self).__init__(trivia)
 
         list.__init__(
@@ -762,18 +788,6 @@ class Array(Item, list):
     def value(self):  # type: () -> list
         return self
 
-    def is_homogeneous(self):  # type: () -> bool
-        if not self:
-            return True
-
-        discriminants = [
-            i.discriminant
-            for i in self._value
-            if not isinstance(i, (Whitespace, Comment))
-        ]
-
-        return len(set(discriminants)) == 1
-
     def multiline(self, multiline):  # type: (bool) -> self
         self._multiline = multiline
 
@@ -792,7 +806,7 @@ class Array(Item, list):
 
         return s
 
-    def append(self, _item):  # type: () -> None
+    def append(self, _item):  # type: (Any) -> None
         if self._value:
             self._value.append(Whitespace(", "))
 
@@ -801,9 +815,6 @@ class Array(Item, list):
 
         self._value.append(it)
 
-        if not self.is_homogeneous():
-            raise ValueError("Array has mixed types elements")
-
     if not PY2:
 
         def clear(self):
@@ -856,7 +867,7 @@ class Array(Item, list):
         return self._value, self._trivia
 
 
-class Table(Item, dict):
+class Table(Item, MutableMapping, dict):
     """
     A table literal.
     """
@@ -869,7 +880,7 @@ class Table(Item, dict):
         is_super_table=False,
         name=None,
         display_name=None,
-    ):  # type: (tomlkit.container.Container, Trivia, bool, ...) -> None
+    ):  # type: (tomlkit.container.Container, Trivia, bool, bool, Optional[str], Optional[str]) -> None
         super(Table, self).__init__(trivia)
 
         self.name = name
@@ -880,7 +891,7 @@ class Table(Item, dict):
 
         for k, v in self._value.body:
             if k is not None:
-                super(Table, self).__setitem__(k.key, v)
+                dict.__setitem__(self, k.key, v)
 
     @property
     def value(self):  # type: () -> tomlkit.container.Container
@@ -914,7 +925,7 @@ class Table(Item, dict):
             key = key.key
 
         if key is not None:
-            super(Table, self).__setitem__(key, _item)
+            dict.__setitem__(self, key, _item)
 
         m = re.match("(?s)^[^ ]*([ ]+).*$", self._trivia.indent)
         if not m:
@@ -941,7 +952,7 @@ class Table(Item, dict):
             key = key.key
 
         if key is not None:
-            super(Table, self).__setitem__(key, _item)
+            dict.__setitem__(self, key, _item)
 
         return self
 
@@ -952,7 +963,7 @@ class Table(Item, dict):
             key = key.key
 
         if key is not None:
-            super(Table, self).__delitem__(key)
+            dict.__delitem__(self, key)
 
         return self
 
@@ -962,8 +973,8 @@ class Table(Item, dict):
     def is_super_table(self):  # type: () -> bool
         return self._is_super_table
 
-    def as_string(self, prefix=None):  # type: () -> str
-        return self._value.as_string(prefix=prefix)
+    def as_string(self):  # type: () -> str
+        return self._value.as_string()
 
     # Helpers
 
@@ -982,42 +993,32 @@ class Table(Item, dict):
 
         return self
 
-    def keys(self):  # type: () -> Generator[str]
-        for k in self._value.keys():
-            yield k
-
-    def values(self):  # type: () -> Generator[Item]
-        for v in self._value.values():
-            yield v
-
-    def items(self):  # type: () -> Generator[Item]
-        for k, v in self._value.items():
-            yield k, v
-
-    def update(self, other):  # type: (Dict) -> None
-        for k, v in other.items():
-            self[k] = v
-
     def get(self, key, default=None):  # type: (Any, Optional[Any]) -> Any
         return self._value.get(key, default)
 
-    def __contains__(self, key):  # type: (Union[Key, str]) -> bool
-        return key in self._value
+    def setdefault(
+        self, key, default=None
+    ):  # type: (Union[Key, str], Any) -> Union[Item, Container]
+        super(Table, self).setdefault(key, default=default)
+
+        return self[key]
 
     def __getitem__(self, key):  # type: (Union[Key, str]) -> Item
         return self._value[key]
 
     def __setitem__(self, key, value):  # type: (Union[Key, str], Any) -> None
+        fix_indent = key not in self
+
         if not isinstance(value, Item):
             value = item(value)
 
         self._value[key] = value
 
         if key is not None:
-            super(Table, self).__setitem__(key, value)
+            dict.__setitem__(self, key, value)
 
         m = re.match("(?s)^[^ ]*([ ]+).*$", self._trivia.indent)
-        if not m:
+        if not m or not fix_indent:
             return
 
         indent = m.group(1)
@@ -1032,11 +1033,14 @@ class Table(Item, dict):
     def __delitem__(self, key):  # type: (Union[Key, str]) -> None
         self.remove(key)
 
-    def __repr__(self):
-        return super(Table, self).__repr__()
+    def __len__(self):  # type: () -> int
+        return len(self._value)
 
-    def __str__(self):
-        return str(self.value)
+    def __iter__(self):  # type: () -> Iterator[str]
+        return iter(self._value)
+
+    def __repr__(self):  # type: () -> str
+        return repr(self._value)
 
     def _getstate(self, protocol=3):
         return (
@@ -1049,7 +1053,7 @@ class Table(Item, dict):
         )
 
 
-class InlineTable(Item, dict):
+class InlineTable(Item, MutableMapping, dict):
     """
     An inline table literal.
     """
@@ -1064,7 +1068,7 @@ class InlineTable(Item, dict):
 
         for k, v in self._value.body:
             if k is not None:
-                super(InlineTable, self).__setitem__(k.key, v)
+                dict.__setitem__(self, k.key, v)
 
     @property
     def discriminant(self):  # type: () -> int
@@ -1093,7 +1097,7 @@ class InlineTable(Item, dict):
             key = key.key
 
         if key is not None:
-            super(InlineTable, self).__setitem__(key, _item)
+            dict.__setitem__(self, key, _item)
 
         return self
 
@@ -1104,7 +1108,7 @@ class InlineTable(Item, dict):
             key = key.key
 
         if key is not None:
-            super(InlineTable, self).__delitem__(key)
+            dict.__delitem__(self, key)
 
         return self
 
@@ -1124,7 +1128,7 @@ class InlineTable(Item, dict):
 
             buf += "{}{}{}{}{}{}".format(
                 v.trivia.indent,
-                k.as_string(),
+                k.as_string() + ("." if k.is_dotted() else ""),
                 k.sep,
                 v.as_string(),
                 v.trivia.comment,
@@ -1140,25 +1144,16 @@ class InlineTable(Item, dict):
 
         return buf
 
-    def keys(self):  # type: () -> Generator[str]
-        for k in self._value.keys():
-            yield k
-
-    def values(self):  # type: () -> Generator[Item]
-        for v in self._value.values():
-            yield v
-
-    def items(self):  # type: () -> Generator[Item]
-        for k, v in self._value.items():
-            yield k, v
-
-    def update(self, other):  # type: (Dict) -> None
-        for k, v in other.items():
-            self[k] = v
-
     def get(self, key, default=None):  # type: (Any, Optional[Any]) -> Any
         return self._value.get(key, default)
 
+    def setdefault(
+        self, key, default=None
+    ):  # type: (Union[Key, str], Any) -> Union[Item, Container]
+        super(InlineTable, self).setdefault(key, default=default)
+
+        return self[key]
+
     def __contains__(self, key):  # type: (Union[Key, str]) -> bool
         return key in self._value
 
@@ -1172,7 +1167,8 @@ class InlineTable(Item, dict):
         self._value[key] = value
 
         if key is not None:
-            super(InlineTable, self).__setitem__(key, value)
+            dict.__setitem__(self, key, value)
+
         if value.trivia.comment:
             value.trivia.comment = ""
 
@@ -1192,8 +1188,14 @@ class InlineTable(Item, dict):
     def __delitem__(self, key):  # type: (Union[Key, str]) -> None
         self.remove(key)
 
+    def __len__(self):  # type: () -> int
+        return len(self._value)
+
+    def __iter__(self):  # type: () -> Iterator[str]
+        return iter(self._value)
+
     def __repr__(self):
-        return super(InlineTable, self).__repr__()
+        return repr(self._value)
 
     def _getstate(self, protocol=3):
         return (self._value, self._trivia)
@@ -1250,7 +1252,7 @@ class AoT(Item, list):
 
     def __init__(
         self, body, name=None, parsed=False
-    ):  # type: (List[Table], Optional[str]) -> None
+    ):  # type: (List[Table], Optional[str], bool) -> None
         self.name = name
         self._body = []
         self._parsed = parsed
@@ -1295,7 +1297,7 @@ class AoT(Item, list):
     def as_string(self):  # type: () -> str
         b = ""
         for table in self._body:
-            b += table.as_string(prefix=self.name)
+            b += table.as_string()
 
         return b
 
